/*
 * Copyright (c) 2016 Intel Corporation.
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <zephyr/logging/log.h>
LOG_MODULE_REGISTER(wifi, CONFIG_LOG_DEFAULT_LEVEL);

#include <zephyr/kernel.h>
#include <errno.h>

#include <zephyr/net/net_if.h>
#include <zephyr/net/wifi_mgmt.h>
#include <zephyr/net/net_event.h>
#include <zephyr/drivers/gpio.h>

#include <zephyr/net/net_ip.h>
#include <zephyr/net/socket.h>
#include <zephyr/net/tls_credentials.h>
#include <zephyr/net/http/client.h>

// Time headers start
#include <zephyr/logging/log.h>
#include <zephyr/net/sntp.h>
#ifdef CONFIG_POSIX_API
#include <arpa/inet.h>
#endif
#include <sys/time.h>
#include <time.h>
#include <zephyr/posix/time.h>
#include <zephyr/sys/timeutil.h>
// Time headers End


#include "http_get.h"
#include "net_private.h"
#include "ca_certificate.h"

#define WIFI_SHELL_MODULE "wifi"

#define WIFI_SHELL_MGMT_EVENTS (NET_EVENT_WIFI_CONNECT_RESULT | \
                                NET_EVENT_WIFI_DISCONNECT_RESULT)

#define MAX_SSID_LEN 32
#define DHCP_TIMEOUT 70
#define CONNECTION_TIMEOUT 100
#define STATUS_POLLING_MS 300

#define HTTP_BUFFER_SIZE 1024

/* 1000 msec = 1 sec */
#define LED_SLEEP_TIME_MS 100

/* The devicetree node identifier for the "led0" alias. */
#define LED0_NODE DT_ALIAS(led0)

/* HTTP server to connect to */
#define HTTP_HOST "google.com"
/* Port to connect to, as string */
#if defined(CONFIG_NET_SOCKETS_SOCKOPT_TLS)
#define HTTP_PORT "443"
#else
#define HTTP_PORT "80"
#endif
/* HTTP path to request */
#define HTTP_PATH "/"


#define SSTRLEN(s) (sizeof(s) - 1)
#define CHECK(r) { if (r == -1) { printf("Error: " #r "\n"); exit(1); } }

#define REQUEST "GET " HTTP_PATH " HTTP/1.0\r\nHost: " HTTP_HOST "\r\n\r\n"

/*
 * A build error on this line means your board is unsupported.
 * See the sample documentation for information on how to fix this.
 */
static const struct gpio_dt_spec led = GPIO_DT_SPEC_GET(LED0_NODE, gpios);

static struct
{
        const struct shell *sh;
        union
        {
                struct
                {
                        uint8_t connected : 1;
                        uint8_t connect_result : 1;
                        uint8_t disconnect_requested : 1;
                        uint8_t _unused : 5;
                };
                uint8_t all;
        };
} context;

static struct net_mgmt_event_callback wifi_shell_mgmt_cb;
static struct net_mgmt_event_callback net_shell_mgmt_cb;


void http_response_cb(struct http_response *response, void *user_data) {
    char *buf = (char *)user_data;
    LOG_INF("Received response: %.*s\n", response->data_len, buf);
}

void toggle_led(void)
{
        int ret;

        if (!device_is_ready(led.port))
        {
                LOG_ERR("LED device is not ready");
                return;
        }

        ret = gpio_pin_configure_dt(&led, GPIO_OUTPUT_ACTIVE);
        if (ret < 0)
        {
                LOG_ERR("Error %d: failed to configure LED pin", ret);
                return;
        }

        while (1)
        {
                if (context.connected)
                {
                        gpio_pin_toggle_dt(&led);
                        k_msleep(LED_SLEEP_TIME_MS);
                }
                else
                {
                        gpio_pin_set_dt(&led, 0);
                        k_msleep(LED_SLEEP_TIME_MS);
                }
        }
}

K_THREAD_DEFINE(led_thread_id, 1024, toggle_led, NULL, NULL, NULL,
                7, 0, 0);

static int cmd_wifi_status(void)
{
        struct net_if *iface = net_if_get_default();
        struct wifi_iface_status status = {0};

        if (net_mgmt(NET_REQUEST_WIFI_IFACE_STATUS, iface, &status,
                     sizeof(struct wifi_iface_status)))
        {
                LOG_INF("Status request failed");

                return -ENOEXEC;
        }

        LOG_INF("==================");
        LOG_INF("State: %s", wifi_state_txt(status.state));

        if (status.state >= WIFI_STATE_ASSOCIATED)
        {
                uint8_t mac_string_buf[sizeof("xx:xx:xx:xx:xx:xx")];

                LOG_INF("Interface Mode: %s",
                        wifi_mode_txt(status.iface_mode));
                LOG_INF("Link Mode: %s",
                        wifi_link_mode_txt(status.link_mode));
                LOG_INF("SSID: %-32s", status.ssid);
                LOG_INF("BSSID: %s",
                        net_sprint_ll_addr_buf(
                            status.bssid, WIFI_MAC_ADDR_LEN,
                            mac_string_buf, sizeof(mac_string_buf)));
                LOG_INF("Band: %s", wifi_band_txt(status.band));
                LOG_INF("Channel: %d", status.channel);
                LOG_INF("Security: %s", wifi_security_txt(status.security));
                LOG_INF("MFP: %s", wifi_mfp_txt(status.mfp));
                LOG_INF("RSSI: %d", status.rssi);
        }
        return 0;
}

static void handle_wifi_connect_result(struct net_mgmt_event_callback *cb)
{
        const struct wifi_status *status =
            (const struct wifi_status *)cb->info;

        if (context.connected)
        {
                return;
        }

        if (status->status)
        {
                LOG_ERR("Connection failed (%d)", status->status);
        }
        else
        {
                LOG_INF("Connected");
                context.connected = true;
        }

        context.connect_result = true;

        cmd_wifi_status();
}

static void handle_wifi_disconnect_result(struct net_mgmt_event_callback *cb)
{
        const struct wifi_status *status =
            (const struct wifi_status *)cb->info;

        if (!context.connected)
        {
                return;
        }

        if (context.disconnect_requested)
        {
                LOG_INF("Disconnection request %s (%d)",
                        status->status ? "failed" : "done",
                        status->status);
                context.disconnect_requested = false;
        }
        else
        {
                LOG_INF("Received Disconnected");
                context.connected = false;
        }

        cmd_wifi_status();
}

static void wifi_mgmt_event_handler(struct net_mgmt_event_callback *cb,
                                    uint32_t mgmt_event, struct net_if *iface)
{
        switch (mgmt_event)
        {
        case NET_EVENT_WIFI_CONNECT_RESULT:
                handle_wifi_connect_result(cb);
                break;
        case NET_EVENT_WIFI_DISCONNECT_RESULT:
                handle_wifi_disconnect_result(cb);
                break;
        default:
                break;
        }
}

static void print_dhcp_ip(struct net_mgmt_event_callback *cb)
{
        /* Get DHCP info from struct net_if_dhcpv4 and print */
        const struct net_if_dhcpv4 *dhcpv4 = cb->info;
        const struct in_addr *addr = &dhcpv4->requested_ip;
        char dhcp_info[128];

        net_addr_ntop(AF_INET, addr, dhcp_info, sizeof(dhcp_info));

        LOG_INF("DHCP IP address: %s", dhcp_info);
}
static void net_mgmt_event_handler(struct net_mgmt_event_callback *cb,
                                   uint32_t mgmt_event, struct net_if *iface)
{
        switch (mgmt_event)
        {
        case NET_EVENT_IPV4_DHCP_BOUND:
                print_dhcp_ip(cb);
                break;
        default:
                break;
        }
}

int bytes_from_str(const char *str, uint8_t *bytes, size_t bytes_len)
{
        size_t i;
        char byte_str[3];

        if (strlen(str) != bytes_len * 2)
        {
                LOG_ERR("Invalid string length: %zu (expected: %d)\n",
                        strlen(str), bytes_len * 2);
                return -EINVAL;
        }

        for (i = 0; i < bytes_len; i++)
        {
                memcpy(byte_str, str + i * 2, 2);
                byte_str[2] = '\0';
                bytes[i] = strtol(byte_str, NULL, 16);
        }

        return 0;
}

void print_time(void)
{
        time_t now = time(NULL);
        struct tm tm2;
        localtime_r(&now, &tm2);
        printf("Current time is now: %s\n", asctime(&tm2));
}

#define SNTP_PORT 123

void set_time(void)
{
        struct sntp_ctx ctx;
        struct sockaddr_in addr;
        struct sntp_time sntp_time = {0};
        int rv;

        /* ipv4 */
        memset(&addr, 0, sizeof(addr));
        addr.sin_family = AF_INET;
        addr.sin_port = htons(SNTP_PORT);

        inet_pton(AF_INET, "132.163.96.1", &addr.sin_addr);

        rv = sntp_init(&ctx, (struct sockaddr *)&addr,
                       sizeof(struct sockaddr_in));
        if (rv < 0)
        {
                LOG_ERR("Failed to init SNTP IPv4 ctx: %d", rv);
                goto end;
        }

        LOG_INF("Sending SNTP IPv4 request...");
        rv = sntp_query(&ctx, 4 * MSEC_PER_SEC, &sntp_time);
        if (rv < 0)
        {
                LOG_ERR("SNTP IPv4 request failed: %d", rv);
                goto end;
        }

        LOG_INF("status: %d", rv);
        LOG_INF("time since Epoch: high word: %u, low word: %u",
                (uint32_t)(sntp_time.seconds >> 32), (uint32_t)sntp_time.seconds);

end:
        sntp_close(&ctx);

        printf("Server time is: %" PRIu64 "\n", sntp_time.seconds);
        struct timespec tp;
        tp.tv_sec = sntp_time.seconds;
        tp.tv_nsec = 0;
        int ret = clock_settime(CLOCK_REALTIME, &tp);
        if (ret != 0)
        {
                printf("Could not set time and date ");
        }
}

static char response[1024];

int main(void)
{
        int ret;

        int i;

        k_sleep(K_SECONDS(5));

        memset(&context, 0, sizeof(context));

        net_mgmt_init_event_callback(&wifi_shell_mgmt_cb,
                                     wifi_mgmt_event_handler,
                                     WIFI_SHELL_MGMT_EVENTS);

        net_mgmt_add_event_callback(&wifi_shell_mgmt_cb);

        net_mgmt_init_event_callback(&net_shell_mgmt_cb,
                                     net_mgmt_event_handler,
                                     NET_EVENT_IPV4_DHCP_BOUND);

        net_mgmt_add_event_callback(&net_shell_mgmt_cb);

        LOG_INF("Starting %s with CPU frequency: %d MHz", CONFIG_BOARD, SystemCoreClock / MHZ(1));
        k_sleep(K_SECONDS(1));

        while (1)
        {
            if (context.connect_result)
            {
                k_sleep(K_SECONDS(2));
                set_time();
                k_sleep(K_SECONDS(2));
                print_time();
                k_sleep(K_SECONDS(5));

                int sock;
                int err;
                struct sockaddr_in *sa;
                struct addrinfo *rp;

                nslookup("google.com", &rp);
                print_addrinfo_results(&rp);

                printk("Connecting to HTTP Server:\n");

                // Load in the TLS cert for HTTPS
                tls_credential_add(CA_CERTIFICATE_TAG, TLS_CREDENTIAL_CA_CERTIFICATE,
                                   ca_certificate, sizeof(ca_certificate));

                
     
                // Create Socket
                sock = socket(rp->ai_family, rp->ai_socktype, IPPROTO_TLS_1_2); // IPPROTO_TCP //IPPROTO_TLS_1_2
                if (sock < 0)
                {
                        printk("Error creating socket\n");
                        return (-1);
                }
                // TLS settings
                sec_tag_t sec_tag_opt[] = {
                    CA_CERTIFICATE_TAG,
                };
                err = setsockopt(sock, SOL_TLS, TLS_SEC_TAG_LIST,
                                 sec_tag_opt, sizeof(sec_tag_opt));
                if (err < 0)
                {
                        LOG_ERR("Failed to set TLS security TAG list. Err: %d", errno);
                        (void)close(sock);
                        return -errno;
                }

                err = setsockopt(sock, SOL_TLS, TLS_HOSTNAME,
                                 HTTP_HOST, sizeof(HTTP_HOST));
                if (err < 0)
                {
                        LOG_ERR("Failed to set TLS_HOSTNAME option. Err: %d", errno);
                        (void)close(sock);
                        return -errno;
                }

                // Iterate through until we get a successful connection
                for (; rp != NULL; rp = rp->ai_next)
                {
                        if (rp->ai_addr->sa_family == AF_INET)
                        {
                                // IPv4 Address
                                sa = (struct sockaddr_in *)rp->ai_addr;
                                sa->sin_port = htons(443);
                                connect(sock, (struct sockaddr *)sa, sizeof(struct sockaddr_in));
                                if (sock > 0)
                                        break;
                        }
                }

                k_sleep(K_SECONDS(5));
        
                http_get(sock, HTTP_HOST, "/");

                k_sleep(K_SECONDS(1));

                close(sock);

                k_sleep(K_FOREVER);
            }

            k_sleep(K_SECONDS(1));
        }

        return 0;
}
